require 'endymion/format'
require 'endymion/field_spec'
require 'endymion/kind_spec'
require 'endymion/query'
require 'endymion/filter'
require 'endymion/sort'

module Endymion
  class API

    attr_reader :datastore
    
    def initialize(datastore)
      @datastore = datastore
    end

    def defentity(kind)
      kind = Format.format_kind(kind)
      kind_spec = KindSpec.new(kind, self)
      yield(kind_spec)
      save_kind_spec(kind_spec)
      pack(kind.to_sym) {|value| pack_record((value || {}).merge(:kind => kind))}
      unpack(kind.to_sym) {|value| unpack_record((value || {}).merge(:kind => kind))}
    end

    def pack(type, &block)
      packers[type] = block
    end

    def packer_defined?(type)
      packers.has_key?(type)
    end

    def unpack(type, &block)
      unpackers[type] = block
    end

    def unpacker_defined?(type)
      unpackers.has_key?(type)
    end

    # Assigns the datastore within the given block
    def with_datastore(name, opts={})
      old_datastore = @datastore
      @datastore = Endymion.new_datastore(name, opts)
      begin
        yield
      rescue
        @datastore = old_datastore
        raise
      end
    end


    #  Saves a record. Any additional parameters will get merged onto the record before it is saved.
    #
    #  Hyperion.save({:kind => :foo})
    #  => {:kind=>"foo", :key=>"<generated key>"}
    #  Hyperion.save({:kind => :foo}, :value => :bar)
    #  => {:kind=>"foo", :value=>:bar, :key=>"<generated key>"}
    def save(record, attrs={})
      save_many([record.merge(attrs || {})]).first
    end

    # Saves multiple records at once.
    def save_many(records)
      unpack_records(datastore.save(pack_records(records)))
    end

    # Creates a record.  While save delegates to a create or update based on the result
    # of new?, this always creates.  This allows you to set the key explicitly
    def create(record)
      create_many([record]).first
    end

    # Create many records at once
    def create_many(records)
      unpack_records(datastore.create(pack_records(records)))
    end

    # Returns true if the record is new (not saved/doesn't have a :key), false otherwise.
    def new?(record)
      !record.has_key?(:key)
    end

    # Retrieves the value associated with the given key from the datastore. nil if it doesn't exist.
    def find_by_key(kind, key)
      unpack_record(datastore.find_by_key(kind, key))
    end

    # Returns all records of the specified kind that match the filters provided.
    #
    #   find_by_kind(:dog) # returns all records with :kind of \"dog\"
    #   find_by_kind(:dog, :filters => [[:name, '=', "Fido"]]) # returns all dogs whos name is Fido
    #   find_by_kind(:dog, :filters => [[:age, '>', 2], [:age, '<', 5]]) # returns all dogs between the age of 2 and 5 (exclusive)
    #   find_by_kind(:dog, :sorts => [[:name, :asc]]) # returns all dogs in alphebetical order of their name
    #   find_by_kind(:dog, :sorts => [[:age, :desc], [:name, :asc]]) # returns all dogs ordered from oldest to youngest, and gos of the same age ordered by name
    #   find_by_kind(:dog, :limit => 10) # returns upto 10 dogs in undefined order
    #   find_by_kind(:dog, :sorts => [[:name, :asc]], :limit => 10) # returns upto the first 10 dogs in alphebetical order of their name
    #   find_by_kind(:dog, :sorts => [[:name, :asc]], :limit => 10, :offset => 10) # returns the second set of 10 dogs in alphebetical order of their name
    #
    # Filter operations and acceptable syntax:
    #   "=" "eq"
    #   "<" "lt"
    #   "<=" "lte"
    #   ">" "gt"
    #   ">=" "gte"
    #   "!=" "not"
    #   "contains?" "contains" "in?" "in"
    #
    # Sort orders and acceptable syntax:
    #   :asc "asc" :ascending "ascending"
    #   :desc "desc" :descending "descending"
    def find_by_kind(kind, args={})
      unpack_records(datastore.find(build_query(kind, args)))
    end

    # Removes the record stored with the given key. Returns nil no matter what.
    def delete_by_key(kind, key)
      datastore.delete_by_key(kind, key)
    end

    # Deletes all records of the specified kind that match the filters provided.
    def delete_by_kind(kind, args={})
      datastore.delete(build_query(kind, args))
    end

    # Counts records of the specified kind that match the filters provided.
    def count_by_kind(kind, args={})
      datastore.count(build_query(kind, args))
    end

    
    
    # NOTE: these methods were marked as private in Hyperion, but were actually
    # accessed from outside, in the FieldSpec class. Temporarily made public until
    # things are cleand up
    def packer_for(type)
      @packers[type]
    end

    def unpacker_for(type)
      @unpackers[type]
    end


    private

    def packers
      @packers ||= {}
    end

    def unpackers
      @unpackers ||= {}
    end

    def build_query(kind, args)
      kind = Format.format_kind(kind)
      filters = build_filters(kind, args[:filters])
      sorts = build_sorts(kind, args[:sorts])
      Query.new(kind, filters, sorts, args[:limit], args[:offset])
    end

    def build_filters(kind, filters)
      (filters || []).map do |(field, operator, value)|
        operator = Format.format_operator(operator)
        packed_field = pack_field(kind, field)
        value = pack_value(kind, field, value)
        Filter.new(packed_field, operator, value)
      end
    end

    def build_sorts(kind, sorts)
      (sorts || []).map do |(field, order)|
        field = pack_field(kind, field)
        order = Format.format_order(order)
        Sort.new(field, order)
      end
    end

    def unpack_records(records)
      records.map do |record|
        unpack_record(record)
      end
    end

    def unpack_record(record)
      if record
        create_entity(record) do |record, entity, field, field_spec|
          value = record[field_spec.db_name]
          entity[field] = field_spec.unpack(value)
          entity
        end
      end
    end

    def pack_records(records)
      records.map do |record|
        pack_record(record)
      end
    end

    def pack_record(record)
      if record
        entity = create_entity(record) do |record, entity, field, field_spec|
          value = record[field]
          entity[field_spec.db_name] = field_spec.pack(value || field_spec.default)
          entity
        end
        update_timestamps(entity)
      end
    end

    def pack_field(kind, field)
      field = Format.format_field(field)
      kind_spec = kind_spec_for(kind)
      return field unless kind_spec
      field_spec = kind_spec.fields[field]
      return field unless field_spec
      return field_spec.db_name
    end

    def pack_value(kind, field, value)
      kind_spec = kind_spec_for(kind)
      return value unless kind_spec
      field_spec = kind_spec.fields[field]
      return value unless field_spec
      return field_spec.pack(value)
    end

    def update_timestamps(record)
      new?(record) ? update_created_at(record) : update_updated_at(record)
    end

    def update_updated_at(record)
      spec = kind_spec_for(record[:kind])
      if spec && spec.fields.include?(:updated_at)
        record[:updated_at] = Time.now
      end
      record
    end

    def update_created_at(record)
      spec = kind_spec_for(record[:kind])
      if spec && spec.fields.include?(:created_at)
        record[:created_at] = Time.now
      end
      record
    end

    def create_entity(record)
      record = Format.format_record(record)
      kind = record[:kind]
      spec = kind_spec_for(kind)
      unless spec
        record
      else
        key = record[:key]
        base_record = {:kind => kind}
        base_record[:key] = key if key
        spec.fields.reduce(base_record) do |new_record, (name, spec)|
          yield(record, new_record, name, spec)
        end
      end
    end

    def kind_spec_for(kind)
      @kind_specs ||= {}
      @kind_specs[kind]
    end

    def save_kind_spec(kind_spec)
      @kind_specs ||= {}
      @kind_specs[kind_spec.kind] = kind_spec
    end







  end
end